@@ -19,8 +19,9 @@ __all__ = ["memoize", "curry", "iscurried",
 from functools import wraps, partial
 from operator import itemgetter
 
-from unpythonic.arity import arities
+from unpythonic.arity import arities, UnknownArity
 from unpythonic.fold import reducel
+from unpythonic.dynscope import dyn, make_dynvar
 
 def memoize(f):
     """Decorator: memoize the function f.
@@ -34,7 +35,7 @@ def memoize(f):
     **CAUTION**: ``f`` must be pure (no side effects, no internal state
     preserved between invocations) for this to make any sense.
     """
-    success, fail = [object() for _ in range(2)]  # sentinels
+    success, fail = [object() for _ in range(2)]
     memo = {}
     @wraps(f)
     def memoized(*args, **kwargs):
@@ -45,11 +46,10 @@ def memoize(f):
             except BaseException as err:
                 result = (fail, err)
             memo[k] = result  # should yell separately if k is not a valid key
-        sentinel, value = memo[k]
-        if sentinel is success:
-            return value
-        else:
+        kind, value = memo[k]
+        if kind is fail:
             raise value
+        return value
     return memoized
 
 #def memoize_simple(f):  # essential idea, without exception handling
@@ -62,6 +62,9 @@ def memoize(f):
 #        return memo[k]
 #    return memoized
 
+make_dynvar(curry_toplevel_passthrough=False)
+make_dynvar(_curry_context=None)
+make_dynvar(_curry_allow_uninspectable=False)  # if True, no-op if not inspectable.
 def curry(f, *args, **kwargs):
     """Decorator: curry the function f.
 
@@ -101,13 +104,9 @@ def curry(f, *args, **kwargs):
 
     **Passthrough**:
 
-    If too many args are given, any extra ones are passed through on the right::
-
-        double = lambda x: 2 * x
-        assert curry(double)(2, "foo") == (4, "foo")
-
-    In passthrough, if an intermediate result is callable it is invoked
-    on the remaining positional args::
+    If too many args are given, any extra ones are passed through on the right.
+    If an intermediate result is callable, it is invoked on the remaining
+    positional args::
 
         map_one = lambda f: (curry(foldr))(composer(cons, to1st(f)), nil)
         assert curry(map_one)(double, ll(1, 2, 3)) == ll(2, 4, 6)
@@ -119,6 +118,19 @@ def curry(f, *args, **kwargs):
     For simplicity, in passthrough, all kwargs are consumed in the first step
     for which too many positional args were supplied.
 
+    By default, if any passed-through positional args are still remaining when
+    the currently top-level curry context exits, ``curry`` raises ``TypeError``,
+    because such usage often indicates a bug.
+
+    This behavior can be locally modified by setting the dynvar
+    ``curry_toplevel_passthrough``::
+
+        with dyn.let(curry_toplevel_passthrough=True):
+            curry(double, 2, "foo") == (4, "foo")
+
+    Because it is a dynvar, it affects all ``curry`` calls in its dynamic extent,
+    including ones inside library functions such as ``composerc`` or ``pipec``.
+
     **Curry itself is curried**:
 
     When invoked as a regular function (not decorator), curry itself is curried.
@@ -129,8 +141,6 @@ def curry(f, *args, **kwargs):
 
     This comboes with passthrough::
 
-        assert curry(double, 2, "foo") == (4, "foo")
-
         mymap = lambda f: curry(foldr, composerc(cons, f), nil)
         add = lambda x, y: x + y
         assert curry(mymap, add, ll(1, 2, 3), ll(4, 5, 6)) == ll(5, 7, 9)
@@ -149,26 +159,39 @@ def curry(f, *args, **kwargs):
         return f
     # TODO: improve: all required name-only args should be present before calling f.
     # Difficult, partial() doesn't remove an already-set kwarg from the signature.
-    min_arity, max_arity = arities(f)
+    try:
+        min_arity, max_arity = arities(f)
+    except UnknownArity:  # likely a builtin
+        if not dyn._curry_allow_uninspectable:  # usual behavior
+            raise
+        # co-operate with the autocurry macro; don't crash on builtins
+        if args or kwargs:
+            return f(*args, **kwargs)
+        return f
     @wraps(f)
     def curried(*args, **kwargs):
-        if len(args) < min_arity:
-            return curry(partial(f, *args, **kwargs))
-        # passthrough on right, like https://github.com/Technologicat/spicy
-        if len(args) > max_arity:
-            now_args, later_args = args[:max_arity], args[max_arity:]
-            now_result = f(*now_args, **kwargs)  # use up all kwargs now
-            if callable(now_result):
-                # curry it now, to sustain the chain in case we still have
-                # too many args.
-                if not iscurried(now_result):
-                    now_result = curry(now_result)
-                return now_result(*later_args)
-            elif isinstance(now_result, (tuple, list)):
-                return tuple(now_result) + later_args
-            else:
+        outerctx = dyn._curry_context
+        with dyn.let(_curry_context=(outerctx, f)):
+            if len(args) < min_arity:
+                return curry(partial(f, *args, **kwargs))
+            # passthrough on right, like https://github.com/Technologicat/spicy
+            if len(args) > max_arity:
+                now_args, later_args = args[:max_arity], args[max_arity:]
+                now_result = f(*now_args, **kwargs)  # use up all kwargs now
+                if callable(now_result):
+                    # curry it now, to sustain the chain in case we have
+                    # too many (or too few) args for it.
+                    if not iscurried(now_result):
+                        now_result = curry(now_result)
+                    return now_result(*later_args)
+                if not (outerctx or dyn.curry_toplevel_passthrough):
+                    raise TypeError("Top-level curry context exited with {:d} arg(s) remaining: {}".format(len(later_args), later_args))
+                # pass through to the curried procedure waiting in outerctx
+                # (e.g. in a curried compose chain)
+                if isinstance(now_result, tuple):
+                    return now_result + later_args
                 return (now_result,) + later_args
-        return f(*args, **kwargs)
+            return f(*args, **kwargs)
     curried._is_curried_function = True  # stash for detection
     # curry itself is curried: if we get args, they're the first step
     if args or kwargs:
@@ -196,16 +219,19 @@ def flip(f):
     return flipped
 
 def rotate(k):
-    """Decorator (factory): cycle positional args of f to the right by k places.
+    """Decorator (factory): cycle positional arg slots of f to the right by k places.
 
     Negative values cycle to the left.
 
-    Note this shifts the incoming argument values, not the formal parameter list!
+    Note this (conceptually) shifts the slots, not the incoming argument values.
 
     **Examples**::
 
-        assert (rotate(1)(identity))(1, 2, 3) == (3, 1, 2)
-        assert (rotate(-1)(identity))(1, 2, 3) == (2, 3, 1)
+        # (a, b, c) -> (b, c, a), so b=1, c=2, a=3 in return (a, b, c)
+        assert (rotate(-1)(identity))(1, 2, 3) == (3, 1, 2)
+
+        # (a, b, c) -> (c, a, b), so c=1, a=2, b=3 in return (a, b, c)
+        assert (rotate(1)(identity))(1, 2, 3) == (2, 3, 1)
     """
     def rotate_k(f):
         @wraps(f)
@@ -213,7 +239,7 @@ def rotate(k):
             n = len(args)
             if not n:
                 raise TypeError("Expected at least one argument")
-            j = k % n  # handle also negative values
+            j = -k % n
             rargs = args[-j:] + args[:-j]
             return f(*rargs, **kwargs)
         return rotated
@@ -389,13 +415,20 @@ def composel1i(iterable):
 def _make_compose(direction):  # "left", "right"
     def compose_two(f, g):
         def composed(*args):
-            a = g(*args)
-            # we could duck-test but this is more predictable for the user
-            # (consider chaining functions that manipulate a generator).
-            if isinstance(a, (list, tuple)):
+            bindings = {}
+            if iscurried(f):
+                # co-operate with curry: provide a top-level curry context
+                # to allow passthrough from the function that is applied first
+                # to the function that is applied second.
+                bindings = {"_curry_context": (dyn._curry_context, composed)}
+            with dyn.let(**bindings):
+                a = g(*args)
+            # we could duck-test, but this is more predictable for the user
+            # (consider chaining functions that manipulate a generator), and
+            # tuple specifically is the pythonic multiple-return-values thing.
+            if isinstance(a, tuple):
                 return f(*a)
-            else:
-                return f(a)
+            return f(a)
         return composed
     if direction == "right":
         compose_two = flip(compose_two)
@@ -411,7 +444,7 @@ def composer(*fs):
 
     This mirrors the standard mathematical convention (f ∘ g)(x) ≡ f(g(x)).
 
-    At each step, if the output from a function is a list or a tuple,
+    At each step, if the output from a function is a tuple,
     it is unpacked to the argument list of the next function. Otherwise,
     we assume the output is intended to be fed to the next function as-is.
 
@@ -643,16 +676,11 @@ def test():
     assert processor(1, 2, 3) == (3, 8, 4)
 
     assert identity(1, 2, 3) == (1, 2, 3)
-    assert (rotate(1)(identity))(1, 2, 3) == (3, 1, 2)
-    assert (rotate(-1)(identity))(1, 2, 3) == (2, 3, 1)
+    assert (rotate(-1)(identity))(1, 2, 3) == (3, 1, 2)
+    assert (rotate(1)(identity))(1, 2, 3) == (2, 3, 1)
 
-    # The inner decorator is applied first to the decorated function, as usual.
-    #
-    # But here the outer one effectively takes effect first, because of the
-    # order in which the decorators get their hands on the incoming arguments.
-    #
-    # So this first flips, then rotates the given arguments:
-    assert flip(rotate(1)(identity))(1, 2, 3) == (1, 3, 2)
+    # inner to outer: (a, b, c) -> (b, c, a) -> (a, c, b)
+    assert flip(rotate(-1)(identity))(1, 2, 3) == (1, 3, 2)
 
     def hello(*args):
         return args
