@@ -99,14 +99,14 @@ def gmemoize(gfunc):
         return _MemoizedGenerator(*memos[k])
     return gmemoized
 
-_success, _fail = [object() for _ in range(2)]  # sentinels (globals, saves indirect via self)
+_success, _fail = [object() for _ in range(2)]  # global saves indirect via self
 class _MemoizedGenerator:
     """Wrapper that manages one memoized sequence. Co-operates with gmemoize."""
     def __init__(self, g, memo, lock):
-        self.memo = memo  # each instance for the same g gets the same memo
         self.g = g
-        self.j = 0  # current position in memo
+        self.memo = memo  # each instance for the same g gets the same memo
         self.lock = lock
+        self.j = 0  # current position in memo
     def __repr__(self):
         return "<_MemoizedGenerator object {} at 0x{:x}>".format(self.g.__name__, id(self))
     def __iter__(self):
@@ -123,12 +123,11 @@ class _MemoizedGenerator:
                 except BaseException as err:  # StopIteration propagates, not a BaseException
                     result = (_fail, err)
                 memo.append(result)
-            sentinel, value = result
+            kind, value = result
             self.j += 1
-            if sentinel is _success:
-                return value
-            else:
+            if kind is _fail:
                 raise value
+            return value
 
 def imemoize(iterable):
     """Memoize an iterable.
@@ -329,14 +328,14 @@ def test():
     def primes():
         yield 2
         for n in count(start=3, step=2):
-            if not any(p != n and n % p == 0 for p in takewhile(lambda x: x*x <= n, primes())):
+            if not any(n % p == 0 for p in takewhile(lambda x: x*x <= n, primes())):
                 yield n
 
     @gmemoize  # <-- the only change (beside the function name)
     def mprimes():
         yield 2
         for n in count(start=3, step=2):
-            if not any(p != n and n % p == 0 for p in takewhile(lambda x: x*x <= n, mprimes())):
+            if not any(n % p == 0 for p in takewhile(lambda x: x*x <= n, mprimes())):
                 yield n
 
     def memo_primes():  # with manually implemented memoization
@@ -345,7 +344,7 @@ def test():
             memo.append(2)
             yield 2
             for n in count(start=3, step=2):
-                if not any(p != n and n % p == 0 for p in takewhile(lambda x: x*x <= n, memo)):
+                if not any(n % p == 0 for p in takewhile(lambda x: x*x <= n, memo)):
                     memo.append(n)
                     yield n
         return manual_mprimes()
