@@ -11,9 +11,8 @@ __all__ = ["begin", "begin0", "lazy_begin", "lazy_begin0",
 from collections import namedtuple
 from unpythonic.env import env
 from unpythonic.misc import call
-from unpythonic.fun import curry
-
-# evil inspect dependency, used only to provide informative error messages.
+from unpythonic.fun import curry, iscurried
+from unpythonic.dynscope import dyn
 from unpythonic.arity import arity_includes, UnknownArity
 
 # sequence side effects in a lambda
@@ -175,9 +174,8 @@ class piped1:
         """
         if f is getvalue:
             return self._x
-        else:
-            cls = self.__class__
-            return cls(f(self._x))  # functional update
+        cls = self.__class__
+        return cls(f(self._x))  # functional update
     def __repr__(self):
         return "<piped1 at 0x{:x}; value {}>".format(id(self), self._x)
 
@@ -237,10 +235,9 @@ class lazy_piped1:
             for g in self._funcs:
                 v = g(v)
             return v
-        else:
-            # just pass on the reference to the original x.
-            cls = self.__class__
-            return cls(x=self._x, _funcs=self._funcs + (f,))
+        # just pass on the reference to the original x.
+        cls = self.__class__
+        return cls(x=self._x, _funcs=self._funcs + (f,))
     def __repr__(self):
         return "<lazy_piped1 at 0x{:x}; initial value now {}, functions {}>".format(id(self), self._x, self._funcs)
 
@@ -250,7 +247,7 @@ def pipe(values0, *bodys):
     The only restriction is that each function must take as many positional
     arguments as the previous one returns.
 
-    At each step, if the output from a function is a list or a tuple,
+    At each step, if the output from a function is a tuple,
     it is unpacked to the argument list of the next function. Otherwise,
     we assume the output is intended to be fed to the next function as-is.
 
@@ -275,12 +272,21 @@ def pipe(values0, *bodys):
         assert (a, b) == (13, "got foo")
     """
     xs = values0
-    for update in bodys:
-        if isinstance(xs, (list, tuple)):
-            xs = update(*xs)
-        else:
-            xs = update(xs)
-    if isinstance(xs, (list, tuple)):
+    n = len(bodys)
+    for k, update in enumerate(bodys):
+        islast = (k == n - 1)
+        bindings = {}
+        if iscurried(update) and not islast:
+            # co-operate with curry: provide a top-level curry context
+            # to allow passthrough from a pipelined function to the next
+            # (except the last one, since it exits the curry context).
+            bindings = {"_curry_context": (dyn._curry_context, update)}
+        with dyn.let(**bindings):
+            if isinstance(xs, tuple):
+                xs = update(*xs)
+            else:
+                xs = update(xs)
+    if isinstance(xs, tuple):
         return xs if len(xs) > 1 else xs[0]
     return xs
 
@@ -314,16 +320,14 @@ class piped:
         xs = self._xs
         if f is getvalue:
             return xs if len(xs) > 1 else xs[0]
+        cls = self.__class__
+        if isinstance(xs, tuple):
+            newxs = f(*xs)
         else:
-            cls = self.__class__
-            if isinstance(xs, (list, tuple)):
-                newxs = f(*xs)
-            else:
-                newxs = f(xs)
-            if isinstance(newxs, (list, tuple)):
-                return cls(*newxs)
-            else:
-                return cls(newxs)
+            newxs = f(xs)
+        if isinstance(newxs, tuple):
+            return cls(*newxs)
+        return cls(newxs)
     def __repr__(self):
         return "<piped at 0x{:x}; values {}>".format(id(self), self._xs)
 
@@ -362,15 +366,14 @@ class lazy_piped:
         if f is runpipe:  # compute now
             vs = self._xs
             for g in self._funcs:
-                if isinstance(vs, (list, tuple)):
+                if isinstance(vs, tuple):
                     vs = g(*vs)
                 else:
                     vs = g(vs)
             return vs if len(vs) > 1 else vs[0]
-        else:
-            # just pass on the references to the original xs.
-            cls = self.__class__
-            return cls(*self._xs, _funcs=self._funcs + (f,))
+        # just pass on the references to the original xs.
+        cls = self.__class__
+        return cls(*self._xs, _funcs=self._funcs + (f,))
     def __repr__(self):
         return "<lazy_piped at 0x{:x}; initial values now {}, functions {}>".format(id(self), self._xs, self._funcs)
 
@@ -466,8 +469,7 @@ def do(*items):
             except UnknownArity:  # well, we tried!
                 pass
             return v(e)
-        else:
-            return v
+        return v
     for item in items:
         if isinstance(item, DoAssign):
             k, v = item
@@ -553,6 +555,15 @@ def test():
                  lambda x, y: (x * 2, y + 1))
     assert (a, b) == (4, 3)
 
+    try:
+        a, b = pipec((1, 2),
+                     lambda x: x + 1,
+                     lambda x: x * 2)
+    except TypeError:
+        pass
+    else:
+        assert False  # error if the curry context exits with args remaining
+
     # optional shell-like syntax
     assert piped1(42) | double | inc | getvalue == 85
 
